<?php

namespace Drupal\custom_search\Controller;

use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Drupal\node\Entity\Node;

class CustomSearchController {

  public function search(Request $request) {
    // Obtener el término de búsqueda desde la URL.
    $query_param = strtolower($request->query->get('q'));
    $type_param = $request->query->get('type', ''); // Obtener el tipo de contenido, por defecto vacío.

    // Obtener parámetros de paginación.
    $page = (int) $request->query->get('page', 0); // Página actual (por defecto 0).
    $items_per_page = (int) $request->query->get('items_per_page', 10); // Elementos por página (por defecto 10).

    // Validar que el parámetro de búsqueda no esté vacío.
    if (empty($query_param)) {
      return new JsonResponse(['error' => 'El parámetro de búsqueda no puede estar vacío.'], 400);
    }

    // Tipos de contenido disponibles.
    $allowed_content_types = ['page', 'proyectos'];

    // Filtrar por tipo de contenido si se especifica en la URL.
    $content_types = $allowed_content_types;
    if (!empty($type_param)) {
      $requested_types = explode(',', $type_param); // Soporte para múltiples tipos separados por comas.
      $content_types = array_intersect($allowed_content_types, $requested_types);

      // Validar que los tipos solicitados sean válidos.
      if (empty($content_types)) {
        return new JsonResponse(['error' => 'El tipo de contenido especificado no es válido.'], 400);
      }
    }

    // Calcular el desplazamiento (offset) para la consulta.
    $offset = $page * $items_per_page;

    // Query de contenido.
    $query = \Drupal::entityQuery('node')
      ->condition('type', $content_types, 'IN')
      ->condition('status', 1) // Solo contenido publicado.
      ->condition('title', '%' . $query_param . '%', 'LIKE')
      ->range($offset, $items_per_page) // Paginación.
      ->accessCheck(TRUE);

    $nids = $query->execute();

    // Cargar nodos.
    $nodes = Node::loadMultiple($nids);
    $results = [];

    foreach ($nodes as $node) {
      $results[] = [
        'id' => $node->id(),
        'title' => $node->getTitle(),
        'type' => $node->getType(),
        'url' => $node->toUrl()->toString(),
        'body' => $node->hasField('body') ? $node->get('body')->value : '', // Agregar el campo Body si existe.
      ];
    }

    // Contar el total de resultados sin paginación.
    $total_query = \Drupal::entityQuery('node')
      ->condition('type', $content_types, 'IN')
      ->condition('status', 1)
      ->condition('title', '%' . $query_param . '%', 'LIKE')
      ->accessCheck(TRUE); // Verificación de acceso.

    $total_results = $total_query->count()->execute();

    // Devolver resultados con metadatos de paginación.
    return new JsonResponse([
      'results' => $results,
      'pagination' => [
        'total' => $total_results,
        'current_page' => $page,
        'items_per_page' => $items_per_page,
        'total_pages' => ceil($total_results / $items_per_page),
      ],
    ]);
  }
}
